import os.path

import pandas as pd
import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt

from scipy.stats import pearsonr
from enum import Enum


class Categories(Enum):
    Dashboard = "dashboard"
    LMS = "lms"
    Games = "games"
    VR = "vr"


class Interviews:
    """
    This class shows data points of every saved interview.
    In this data frame is also saved how much time was taken for the questionnaire and some additional variables as you
    can see here: https://www.soscisurvey.de/help/doku.php/en:results:variables
    """

    def __init__(self) -> None:
        super().__init__()
        self.df = pd.read_csv('../data/data_ueq.csv', delimiter=';', quotechar='"', encoding='utf-16')

    def __str__(self) -> str:
        return f'{self.df.shape[0]} interviews'

    def get_items_per_product_category(self, product_category_items):
        return self.df[product_category_items]

    def get_filtered(self, all_categories):
        if isinstance(all_categories, ProductCategories):
            cat_games = all_categories.get_items(Categories.Games)
            cat_lms = all_categories.get_items(Categories.LMS)
            # remove interviews which are not complete
            print(cat_games)
            print(self.df)
            unwanted_interviews = self.df[self.df[cat_games[0]].isnull() | self.df[cat_lms[0]].isnull()]
            print(f'{unwanted_interviews.shape[0]} interviews have not completed the questionnaire.')
            df_nonnull_interviews = self.df.drop(unwanted_interviews.index)
            # remove interviews which are, based on indicators, not well answered
            df_too_fast_interviews = df_nonnull_interviews[
                (df_nonnull_interviews['TIME_RSI'] > 2.0) | (df_nonnull_interviews['DEG_TIME'] > 150)]
            print(f'{df_too_fast_interviews.shape[0]} interviews have completed too fast or too weird.')
            result = df_nonnull_interviews.drop(df_too_fast_interviews.index)
            result = result.rename(columns={"EX08_21": "EX08_01", "EX08_22": "EX08_02",
                                                      "EX08_23": "EX08_03", "EX08_24": "EX08_04",
                                                      "EX08_25": "EX08_05", "EX08_26": "EX08_06",
                                                      "EX08_27": "EX08_07", "EX08_28": "EX08_08",
                                                      "EX08_29": "EX08_09", "EX08_30": "EX08_10",
                                                      "EX08_31": "EX08_11", "EX08_32": "EX08_12",
                                                      "EX08_33": "EX08_13", "EX08_34": "EX08_14",
                                                      "EX08_35": "EX08_15", "EX08_36": "EX08_16",
                                                      "EX08_37": "EX08_17", "EX08_38": "EX08_18",
                                                      "EX08_39": "EX08_19", "EX08_40": "EX08_20"
                                                      })
            print(f'{result.shape[0]} filtered interviews.')
            return result


class Variables:
    """
    Gives an explanation of every variable used in the questionnaire.
    """

    def __init__(self) -> None:
        super().__init__()
        self.df = pd.read_csv('../data/variables_ueq.csv', delimiter=';', quotechar='"', encoding='utf-16')

    def get_groups(self):
        return self.df[['VAR', 'LABEL']][7:12]


class Values:
    """
    Gives an explanation of every given value in the interviews
    """

    def __init__(self) -> None:
        super().__init__()
        self.df = pd.read_csv('../data/values_ueq.csv', delimiter=';', quotechar='"', encoding='utf-16')


class ProductCategories:
    """
    Product categories are the core of this evaluation.
    The goal is to evaluate which items are important for the product categories.
    In the created dict in items every product category is connected with the used code and names of their items.
    """

    def __init__(self) -> None:
        super().__init__()
        self.items = {
            Categories.Dashboard: {
                'code': 'EX01'
            },
            Categories.Games: {
                'code': 'EX05'
            },
            Categories.LMS: {
                'code': 'EX07'
            },
            Categories.VR: {
                'code': 'EX08'
            },
        }

        # create the tokens to access every item in a product category
        for key in self.items.keys():
            self.items[key]['items'] = [self.items[key]['code'] + "_" + "{:02d}".format(i) for i in range(1, 21)]

    def get_items(self, key):
        return self.items[key]['items']


def save_plot(plot, path):
    dir_path = os.path.abspath('../results/')
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)
    plot.savefig(fname=os.path.join(dir_path, path + '.png'), dpi='figure', format='png')


def replace_cryptic_names(columns):
    new_columns = list(columns)
    for i in range(len(columns)):
        if "_01" in new_columns[i]:
            new_columns[i] = 'Inhaltsqualität'
        elif "_02" in new_columns[i]:
            new_columns[i] = 'Anpassbarkeit'
        elif "_03" in new_columns[i]:
            new_columns[i] = 'Durchschaubarkeit'
        elif "_04" in new_columns[i]:
            new_columns[i] = 'Effizienz'
        elif "_05" in new_columns[i]:
            new_columns[i] = 'Immersion'
        elif "_06" in new_columns[i]:
            new_columns[i] = 'Intuitive Bedienung'
        elif "_07" in new_columns[i]:
            new_columns[i] = 'Nützlichkeit'
        elif "_08" in new_columns[i]:
            new_columns[i] = 'Originalität'
        elif "_09" in new_columns[i]:
            new_columns[i] = 'Schönheit'
        elif "_10" in new_columns[i]:
            new_columns[i] = 'Identität'
        elif "_11" in new_columns[i]:
            new_columns[i] = 'Steuerbarkeit'
        elif "_12" in new_columns[i]:
            new_columns[i] = 'Stimulation'
        elif "_13" in new_columns[i]:
            new_columns[i] = 'Übersichtlichkeit'
        elif "_14" in new_columns[i]:
            new_columns[i] = 'Verbundenheit'
        elif "_15" in new_columns[i]:
            new_columns[i] = 'Vertrauen'
        elif "_16" in new_columns[i]:
            new_columns[i] = 'Wertigkeit'
        elif "_17" in new_columns[i]:
            new_columns[i] = 'Attraktivität'
        elif "_18" in new_columns[i]:
            new_columns[i] = 'Haptik'
        elif "_19" in new_columns[i]:
            new_columns[i] = 'Akustik'
        elif "_20" in new_columns[i]:
            new_columns[i] = 'Vertrauenswürdigkeit'

    return new_columns


def filter_columns_by_value(df_category, value):
    filtered_columns = df_category.columns[value]
    df_category_area = df_category[filtered_columns]
    return df_category_area


class Plotter:

    def __init__(self, df, pc) -> None:
        super().__init__()
        self.df = df
        self.pc = pc

    def plot_groups(self, variable_labels):
        df_groups = self.df[['DE01_01', 'DE01_02', 'DE01_03', 'DE01_04', 'DE01_05']] - 1
        print(variable_labels)

        df_groups.columns = ['Stud', 'Doz', 'WiMi', 'TeVe', 'NU']
        sns.set_theme(style="whitegrid")
        ax = sns.barplot(data=df_groups, palette="Blues_d", estimator=np.count_nonzero, ci=None)
        ax.set(title='Statusgruppen')
        sns.despine(left=True)
        save_plot(plt, 'distributions_by_groups')
        plt.show()

    def plot_sex(self):
        df_sex = self.df['DE02'] \
            .replace(1, 'male') \
            .replace(2, 'female') \
            .replace(3, 'diverse') \
            .replace(-1, 'not mentioned')
        ax = sns.countplot(x=df_sex, palette="Blues_d")
        ax.set(title='Geschlecht')
        print(df_sex.value_counts(normalize=True))
        save_plot(plt, 'distribution_by_gender')
        plt.show()

    def plot_faculty(self):
        df_faculty = self.df['DE03'] \
            .replace(1, 'Jur') \
            .replace(2, 'Phil') \
            .replace(3, 'HuWi') \
            .replace(4, 'WiSo') \
            .replace(5, 'MathNat') \
            .replace(6, 'Ges') \
            .replace(7, 'HPI') \
            .replace(8, 'ZE') \
            .replace(9, 'Other') \
            .replace(-1, 'kA') \
            .replace(-9, 'nB')
        ax = sns.countplot(x=df_faculty, palette="Blues_d")
        ax.set(title='Fakultät')
        print(df_faculty.value_counts(normalize=True))
        save_plot(plt, 'distribution_by_faculty')
        plt.show()

    def plot_item(self, category_title, area, max_count=20):
        category_items = self.pc.get_items(category_title)
        df_category = self.df[category_items]
        if area == 'top':
            df_category_area = filter_columns_by_value(df_category, df_category.mean() >= 5)
        elif area == 'mid':
            df_flc = filter_columns_by_value(df_category, df_category.mean() < 5)
            df_category_area = filter_columns_by_value(df_flc, df_flc.mean() >= 3)
        elif area == 'low':
            df_category_area = filter_columns_by_value(df_category, df_category.mean() < 3)
        else:
            df_category_area = df_category
        print('------------- ' + area)
        df_category_area.columns = replace_cryptic_names(df_category_area.columns)
        df_category_area = df_category_area.reindex(df_category_area.mean().sort_values(ascending=False).index, axis=1)
        df_category_area = df_category_area.iloc[:, : max_count]
        print(df_category_area.columns)
        ax = sns.barplot(data=df_category_area, palette="Blues_d", orient='h', ci="sd")
        ax.set(title=category_title.value + ' - ' + area)
        sns.despine(left=True)
        save_plot(plt, 'items_' + category_title.value + '_' + area)
        plt.show()

    def plot_boxes(self, category_title, max_count=5):
        df_category_area = self.df[self.pc.get_items(category_title)]
        df_category_area.columns = replace_cryptic_names(df_category_area.columns)

        # choose 5 most relevant scales (highest mean)
        df_category_area = df_category_area.reindex(df_category_area.mean().sort_values(ascending=False).index, axis=1)
        df_category_area = df_category_area.iloc[:, : max_count]

        # make a nice plot surrounding
        sns.set_theme(style="whitegrid", palette="pastel")
        fig = plt.figure()
        fig.suptitle(category_title.value, fontsize=14, fontweight='bold')

        # plot the plot
        ax = fig.add_subplot(111)
        ax.boxplot(df_category_area)
        ax.set_xticks([tick + 1 for tick in range(len(df_category_area.columns))],
                      labels=df_category_area.columns, rotation=10)
        save_plot(plt, 'box_' + category_title.value)
        plt.show()


class Statistics:

    def __init__(self, df, pc) -> None:
        super().__init__()
        self.df = df
        self.pc = pc

    def calc_corr(self):
        df_means_dashboard = pd.DataFrame(
            {"dashboard": self.df[self.pc.get_items(Categories.Dashboard)].mean()})
        df_means_lms = pd.DataFrame({"lms": self.df[self.pc.get_items(Categories.LMS)].mean()})
        df_means_games = pd.DataFrame({"games": self.df[self.pc.get_items(Categories.Games)].mean()})
        df_means_vr = pd.DataFrame({"vr": self.df[self.pc.get_items(Categories.VR)].mean()})
        df_means_dashboard.index = replace_cryptic_names(df_means_dashboard.index)
        df_means_lms.index = replace_cryptic_names(df_means_lms.index)
        df_means_games.index = replace_cryptic_names(df_means_games.index)
        df_means_vr.index = replace_cryptic_names(df_means_vr.index)
        df_all_means = pd.concat([df_means_lms, df_means_dashboard, df_means_games, df_means_vr], axis=1)

        result = df_all_means.corr()
        pval = df_all_means.corr(method=lambda x, y: pearsonr(x, y)[1]) - np.eye(*result.shape)
        p = pval.applymap(lambda x: ''.join(['*' for t in [.05, .01, .001] if x <= t]))
        return result.round(2).astype(str) + p

    def welch(self):

        return

def setup_dataframes():
    return Interviews(), Variables(), Values(), ProductCategories()


if __name__ == '__main__':
    interview, variables, values, product_categories = setup_dataframes()
    example = Categories.Dashboard
    print(interview)
    df_filtered = interview.get_filtered(product_categories)
    plotter = Plotter(df_filtered, product_categories)
    statistics = Statistics(df_filtered, product_categories)
    plotter.plot_groups(variables.get_groups())
    plotter.plot_sex()
    plotter.plot_faculty()
    plotter.plot_item(example, 'top')
    plotter.plot_item(example, 'mid')
    plotter.plot_item(example, 'low')
    plotter.plot_boxes(example, 5)
    print(statistics.calc_corr())
